// Copyright 2024 schukai GmbH
// SPDX-License-Identifier: AGPL-3.0

package jobqueue

import (
	"errors"
	"gorm.io/gorm"
	"strings"
	"time"
)

func (s *JobSyncer) CheckAndSaveOrUpdate(job GenericJob) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	if s.manager == nil || s.manager.database == nil {
		return ErrNoDatabaseConnection
	}

	permJob := job.GetPersistence()
	db := s.manager.database

	var existing JobPersistence
	result := db.Where("id = ?", permJob.ID).First(&existing)

	if result.Error == nil {
		return updateJob(job, db)
	}

	if errors.Is(result.Error, gorm.ErrRecordNotFound) {
		return saveJob(job, db)
	}

	return result.Error
}

func (s *JobSyncer) DeleteJob(job GenericJob) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	if s.manager == nil || s.manager.database == nil {
		return ErrNoDatabaseConnection
	}

	return s.manager.database.Transaction(func(tx *gorm.DB) error {
		permJob := job.GetPersistence()
		if err := tx.Where("job_id = ?", permJob.GetID()).Delete(&JobLog{}).Error; err != nil {
			return err
		}
		if err := tx.Where("job_id = ?", permJob.GetID()).Delete(&JobStats{}).Error; err != nil {
			return err
		}
		if err := tx.Delete(&permJob).Error; err != nil {
			return err
		}
		return nil
	})
}

func (s *JobSyncer) ResetLogs(job GenericJob) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	if s.manager == nil || s.manager.database == nil {
		return ErrNoDatabaseConnection
	}

	return s.manager.database.Transaction(func(tx *gorm.DB) error {
		permJob := job.GetPersistence()
		if err := tx.Unscoped().Where("job_id = ?", permJob.GetID()).Delete(&JobLog{}).Error; err != nil {
			return err
		}

		return nil
	})
}

func (s *JobSyncer) ResetStats(job GenericJob) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	if s.manager == nil || s.manager.database == nil {
		return ErrNoDatabaseConnection
	}

	job.ResetStats()
	if s.manager == nil || s.manager.database == nil {
		return ErrNoDatabaseConnection
	}
	return saveJob(job, s.manager.database)

}

func (s *JobSyncer) UpdateJob(job GenericJob) error {
	s.mu.Lock()
	defer s.mu.Unlock()
	if s.manager == nil || s.manager.database == nil {
		return ErrNoDatabaseConnection
	}
	return updateJob(job, s.manager.database)
}

func updateJob(job GenericJob, db *gorm.DB) error {

	if db == nil {
		return ErrNoDatabaseConnection
	}

	permJob := job.GetPersistence()

	maxRetries := 3
	var attempt int
	for attempt = 0; attempt < maxRetries; attempt++ {
		err := update(&permJob, db)
		if err != nil {
			if strings.Contains(err.Error(), "lock") {
				time.Sleep(time.Millisecond * 100)
				continue
			}
			return err
		}
		break
	}

	if attempt == maxRetries {
		return ErrMaxRetriesReached
	}

	return nil

}

func update(job *JobPersistence, db *gorm.DB) error {

	if db == nil {
		return ErrNoDatabaseConnection
	}

	permJob := job.GetPersistence()

	return db.Transaction(func(tx *gorm.DB) error {

		if permJob.Stats != (JobStats{}) {
			if err := tx.Model(&permJob).Select("Stats").Updates(map[string]interface{}{"stats": permJob.Stats}).Error; err != nil {
				return err
			}
		}

		for i, _ := range job.Logs {
			job.Logs[i].LogID = 0
			_ = tx.Create(&job.Logs[i])
			// no error handling, if it fails, it fails
		}

		return nil
	})
}

func (s *JobSyncer) SaveJob(job GenericJob) error {
	s.mu.Lock()
	defer s.mu.Unlock()
	if s.manager == nil || s.manager.database == nil {
		return ErrNoDatabaseConnection
	}
	return saveJob(job, s.manager.database)
}

func save(job *JobPersistence, db *gorm.DB) error {

	if db == nil {
		return ErrNoDatabaseConnection
	}

	return db.Transaction(func(tx *gorm.DB) error {

		if err := tx.Save(job).Error; err != nil {
			return err
		}

		if job.Stats != (JobStats{}) {
			job.Stats.JobID = job.ID
			if err := tx.Save(&job.Stats).Error; err != nil {
				return err
			}
		}

		for i, _ := range job.Logs {
			job.Logs[i].LogID = 0
			_ = tx.Create(&job.Logs[i])
			// no error handling, if it fails, it fails
		}

		return nil
	})
}

func saveJob(job GenericJob, db *gorm.DB) error {

	if db == nil {
		return ErrNoDatabaseConnection
	}

	permJob := job.GetPersistence()

	maxRetries := 3
	var attempt int
	for attempt = 0; attempt < maxRetries; attempt++ {
		err := save(&permJob, db)
		if err != nil {
			if strings.Contains(err.Error(), "lock") {
				time.Sleep(time.Millisecond * 100)
				continue
			}
			return err
		}
		break
	}

	if attempt == maxRetries {
		return ErrMaxRetriesReached
	}

	return nil

}
